#
# node.py - Node and NodeList.
# Copyright (c) 2005 shinGETsu Project.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: node.py,v 1.42 2005/10/16 03:48:20 fuktommy Exp $
#

import gzip
import random
import re
import os
import socket
import StringIO
import sys
import time
import urllib

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading

import config

__version__ = "$Revision: 1.42 $"
__all__ = ["Node", "ListFile", "NodeList", "SearchList"]

urllib.URLopener.version = config.version

class SakuAgent(urllib.FancyURLopener):
    def __init__(self):
        self.version = config.version
        urllib.FancyURLopener.__init__(self, proxies={})
        self.addheader("Accept-Encoding", "gzip")

agent = SakuAgent()


class Broadcast(_threading.Thread):
    def __init__(self, msg, cache):
        _threading.Thread.__init__(self)
        self.msg = msg
        self.cache = cache

    def run(self):
        nodelist = NodeList()
        for node in self.cache.node:
            if (node in nodelist) or node.ping():
                node.talk(self.msg)
            else:
                self.cache.node.remove(node)
                self.cache.node.sync()
        for node in nodelist:
            if node not in self.cache.node:
                node.talk(self.msg)

class Node:

    """One unit for P2P."""

    nodestr = None

    def __init__(self, nodestr=None, host="", port="", path=""):
        if nodestr is not None:
            nodestr = re.sub(r"[\r\n]*$", "", nodestr)
            self.nodestr = re.sub(r"\+", "/", nodestr)
        else:
            self.nodestr = host + ":" + str(port) + re.sub(r"\+", "/", path)

    def __str__(self):
        return self.nodestr

    def __gt__(self, y):
        return str(self) > str(y)

    def __lt__(self, y):
        return str(self) < str(y)

    def __eq__(self, y):
        return str(self) == str(y)

    def __ne__(self, y):
        return str(self) != str(y)

    def toxstring(self):
        """Return string, ``/'' is replaced with ``+''."""
        return self.nodestr.replace("/", "+")

    def talk(self, message):
        """Connect other node."""
        if not message.startswith("/"):
            message = "/" + message
        if message.startswith("/get"):
            socket.setdefaulttimeout(config.get_timeout)
        else:
            socket.setdefaulttimeout(config.timeout)

        try:
            sys.stderr.write("talk: http://" + str(self) + message + "\n")
            f = agent.open("http://%s%s" % (self, message))
            buf = f.read()
            f.close()
        except (IOError, AttributeError, socket.error, socket.timeout), err:
            sys.stderr.write("%s%s: %s\n" % (self, message, err))
            return []
        
        if f.info().getheader("Content-Encoding", "") == "gzip":
            sio = StringIO.StringIO(buf)
            try:
                try:
                    gz = gzip.GzipFile(fileobj=sio)
                    buf = gz.read()
                finally:
                    gz.close()
            except IOError, err:
                sys.stderr.write("%s%s: %s\n" % (str(self), message, str(err)))
        return buf.splitlines()

    def ping(self):
        buf = self.talk("/ping")
        return buf and (buf[0] == "PONG")

    def join(self):
        """Connect."""
        port = config.port
        path = config.server
        path = path.replace("/", "+")
        name = config.dnsname
        buf = self.talk("/join/" + name + ":" + str(port) + path)

        welcome = buf and (buf[0] == "WELCOME")
        if len(buf) > 1:
            extnode = Node(buf[1])
        else:
            extnode = None
	return (welcome, extnode)

    def get_node(self):
        buf = self.talk("/node")
        if buf and re.search(r"\d+/[^: ]+$", buf[0]):
            return Node(buf[0])
        else:
            return None

    def bye(self):
        """Unconnect."""
        port = config.port
        path = config.server
        path = path.replace("/", "+")
        name = config.dnsname
        buf = self.talk("/bye/" + name + ":" + str(port) + path)
        return buf and (buf[0] == "BYEBYE")

# End of Node


class ListFile(list):

    """File includes list.

    One element par one line.
    """

    def __init__(self, filename):
        """Load list form file."""
        self.filepath = filename
        try:
            f = file(self.filepath)
            for line in f:
                self.append(Node(line))
            f.close()
            random.shuffle(self)
        except IOError:
            #sys.stderr.write(self.filepath + ": IOError\n")
            pass

    def sync(self):
        """Save list to file."""
        self.sort()
        try:
            f = file(self.filepath, "wb")
            tmp = ""
            for n in self:
                if n != tmp:
                    f.write(str(n) + "\n")
                    tmp = n
            f.close()
        except IOError:
            sys.stderr.write(self.filepath + ": Permission denied\n")

    def random(self):
        return random.choice(self)

    def remove(self, value):
        try:
            list.remove(self, value)
        except ValueError:
            pass

# End of ListFile

class NodeList(ListFile):

    """Node list."""

    def __init__(self):
        ListFile.__init__(self, config.node)

    def myself(self):
        """Who am I."""
        port = config.port
        path = config.server
        dnsname = config.dnsname
        if dnsname != "":
            return Node(host=dnsname, port=port, path=path)
        for n in self:
            buf = n.talk("/ping")
            if (len(buf) > 1) and (buf[0] == "PONG"):
                return Node(host=buf[1], port=port, path=path)
	return None

    def pingall(self):
        """Ping all nodes."""
        for n in self[0:]:
            if not n.ping():
                self.remove(n)

    def join(self, node):
        """Join network."""
        port = config.port
        count = 0
        while count < config.retry:
            count += 1
            (welcome, extnode) = node.join()
            if welcome and (extnode is None):
                self.append(node)
                break
            elif welcome:
                self.append(node)
                node = extnode
            else:
                self.remove(node)
                break

    def init(self):
        """Connect initial node."""
        port = config.port
        for i in config.init_node:
            inode = Node(i)
            if inode.ping():
                self.join(inode)
                break
	myself = self.myself();
        if myself and (myself in self):
            self.remove(myself)
        if len(self) == 0:
            return

        done = {}
        for i in self[:]:
            if i != inode:
                self.join(i)
            done[str(i)] = 1

        while True:
            n = random.choice(self)
            new = n.get_node()
            if (new is not None) and (str(new) not in done):
                self.join(new)
                done[str(new)] = 1
            done[str(n)] = done.get(str(n), 1) + 1
            if (done[str(n)] > config.retry) or (len(self) >= config.nodes):
                break

        if len(self) >= config.nodes:
            inode.bye()
            self.remove(inode)
        elif len(self) <= 1:
            sys.stderr.write("Warning: Few linked nodes.\n")
        while len(self) > config.nodes:
            n = self.random()
            n.bye()
            self.remove(n)

    def tell_update(self, cache, stamp="", id="", node=None):
        """Tell update other nodes.

        If node is None, node is myself.
        """
        if node is None:
            tellstr = ":" + \
                      str(config.port) + \
                      config.server.replace("/", "+")
        else:
            tellstr = node.toxstring()

        arg = "/".join(("", "update", cache.datfile, str(stamp), id, tellstr))

        if "light_cgi" in config.flags:
            broadcast = Broadcast(arg, cache)
            broadcast.start()
        else:
            cmdline = " ".join((sys.executable, config.broadcast, arg))
            files = os.popen2(cmdline)
            for f in files:
                f.close()

# End of NodeList


class SearchList(ListFile):

    """List of existing nodes."""

    def __init__(self):
        ListFile.__init__(self, config.search)

    def search(self, cache=None, myself=None):
        """Search node which has the file."""
        nodelist = NodeList()
        random.shuffle(self)
        count = 0
        for n in self:
            if myself and (n == myself):
                continue
            elif str(n) in config.ignore_search_node:
                continue
            count += 1
            buf = n.talk("/have/" + cache.datfile)
            if buf and (buf[0] == "YES"):
                self.sync()
                return n
            elif buf and (buf[0] == "NO"):
                pass
            elif not n.ping():
                self.remove(n)
                cache.node.remove(n)
            if (count > config.search_depth):
                break
        self.sync()
        if count <= 1:
            sys.stderr.write("Warning: Search nodes are null.\n")
        return None

# End of SearchList
